/** \file player.c Actual player program for 8k MLC.
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <vs1000.h>
#include <audio.h>
#include <mappertiny.h>
#include <minifat.h>
#include <codec.h>
#include <vsNand.h>
#include <player.h>
#include <usblowlib.h>

#include <dev1000.h>
#include <romfont.h>

/*
  Defines:

  USE_DISPLAY
  - Use the default OLED display of the VS1000 developer board.
    The LCD routines from vs1000 developer libarary are used
    (dev1000.h /libdev1000.a). If you use another display, you need
    to build your own initialization string.
  - During playback the display is updated with various information:
    + play position
    + play/pause icon
    + EarSpeaker level, if enabled
    + Volume level
    + shuffle icon, if enabled
    + remaining playtime (if enabled with GET_PLAYTIME)
    + file name
  BACKLIGHT_DIM
  - The display is dimmed after BACKLIGHT_DIM ticks have passed from
    the last key press. If BACKLIGHT_DIM is not defined, this feature
    is removed.
  QUIET_CANCEL
  - Pause mode can be enabled at any time, and the decoder can have
    samples waiting to be played. If you change songs directly from
    pause mode, these waiting samples will be played before the
    decoding of the current file is cancelled. This may be distracting.
    With this define the output volume is lowered while the remaining
    samples are playing and volume is restored for the next song.
  GET_PLAYTIME
  - Enables and uses a routine to calculate the remaining play time
    of an Ogg Vorbis file. Or actually the number of samples in the
    Ogg Vorbis file. The number of samples is converted to seconds
    by dividing it by cs.sampleRate, which is valid when the file
    is being played. (You can also read the samplerate from the
    right offset at the beginning of the Ogg Vorbis file.)
 */
#define USE_DISPLAY
#define BACKLIGHT_DIM (10*TIMER_TICKS) /*dim backlight of OLED after 5 secs*/
#define QUIET_CANCEL /*quiet exit when next or prev pressed from pause mode*/
#define GET_PLAYTIME

extern struct FsNandPhys fsNandPhys;
extern struct FsPhysical *ph;
extern struct FsMapper *map;
extern struct Codec *cod;
extern struct CodecServices cs;
extern u_int16 codecVorbis[];

//extern struct FsMapperFlash fsMapFlash; /*32 words*/
extern struct FsMapperTiny fsMapTiny;

#ifdef GET_PLAYTIME
u_int32 lastSample;
void GetPlayTime(void) {
#define SEEK_LEN 5000 /*max 8192*/
    Seek(minifatInfo.fileSize-SEEK_LEN);
    ReadFile(mallocAreaX, 0, SEEK_LEN);
    lastSample = 0;
    /* 
      "OggS" 4
      '\0'   1 stream structure version
             1 header type flag 1:continue packet, 2:first page, 4:last page
      6-13 absoluteGranulePositionLO..HI
    */
    {
	register int i;
	for(i=SEEK_LEN/2-5;i>=0;i--) {
	    if (mallocAreaX[i] == (('O'<<8)|'g') &&
		mallocAreaX[i+1] == ('g'<<8)|'S') {
		/*TODO: handle larger than 32-bit */
		lastSample = SwapWord(mallocAreaX[i+3]) |
		    ((u_int32)SwapWord(mallocAreaX[i+4]) << 16);
		break;
	    } else if ((mallocAreaX[i] & 255) == 'O' &&
		       mallocAreaX[i+1] == (('g'<<8)|'g')) {
		/*TODO: handle larger than 32-bit */
		lastSample = 
		    (mallocAreaX[i+3] & 255) |
		    ((u_int32)SwapWord(mallocAreaX[i+4]) << 8) |
		    ((u_int32)(mallocAreaX[i+5] & 0xff00) << 16);
		break;
	    }
	}
    }
    Seek(0);
}
#endif

#ifdef USE_DISPLAY
#ifdef BACKLIGHT_DIM
u_int32 backLightTime;
void MyKeyEventHandler(enum keyEvent event) {
    RealKeyEventHandler(event);

    LcdContrast(0x40);
    backLightTime = timeCount;
    SetHookFunction((u_int16)USBSuspend, NullHook); /* disable suspend */
}
#endif

void PlayerDisplay(void) { /*268 words*/ /* does not clear it */
//    PERIP(GPIO1_ODATA) |= LED1;

    /* If no files, do not try to print out info. */
    if (player.totalFiles == 0)
	return;

#if 1 /* Playing time, for audio book needs hours too! (use 32-bit division) */
    {
	register u_int32 t = cs.playTimeSeconds;
	register u_int16 h = t / 3600;
	LcdLocate(0, 0);
	if (t == -1L) {
	    LcdPutPackedX("\p  ", 1/2, 42);
	} else {
	    LcdPutChar(h%10 +SYMBOL_DIGIT);
	    t -= (u_int32)h * 3600;
	    LcdPutChar(':');
	    LcdPutNum((u_int16)t / 60);
	    LcdPutChar(':');
	    LcdPutNum((u_int16)t % 60);
	}
    }
#endif

    LcdLocate(42, 0);
    LcdPutChar(player.pauseOn ? SYMBOL_PAUSE : SYMBOL_PLAY);
#if 1
    /* earspeaker -- gfx*/
    {
	const u_int16 * const earGfx[4] = {
	    "\x0000\x0000\x0000\x0000\x0000\x0000\x0000\x0000",
	    "\x8292\x926c\x00fe\x8282\x4438\x00c0\x0000\x0000",
	    "\x8292\x926c\x00fe\x8282\x4438\x00c0\x00f8\x0000",
	    "\x8292\x926c\x00fe\x8282\x4438\x00c0\x00f8\x00ff",
	};
	LcdPutGfx(earGfx[earSpeakerReg/18000U], 8);
    }
#endif

#if 1 /* volume */
    LcdLocate(70, 0);
    LcdPutChar(SYMBOL_RIGHTSPK);
    {
	//register int len = 12 - (24 + player.volume)/4; /* 0..xx */
	register int len = 6 - player.volume / 4; /* 0..xx */
	if (len > 0) {
	    LcdPutGfx("\x8000\xc000\xe000\xf000\xf800\xfc00\xfe00\xff80"
		      "\xffc0\xffe0\xfff0\xfff8\xfffc\xfffe\xffff\xffff",
		      len);
	}
	LcdPutPackedX("\p", 0, 69+31);
    }
#endif

#if 1
    LcdLocate(101, 0);
    if (player.randomOn) {
	LcdPutGfx("\x0042\x4242\x4224\x1008\x2442\x4242\xe742", 7);
    }
    LcdPutPackedX("", 0, 132);
#endif

#if 0
    LcdLocate(101, 0);
    LcdPutData(0xff);
    LcdPutData(0x81);
    screen.invertMask = 0xbd;
    LcdPutPackedX("", 0, (100-12)+battery.level);
    screen.invertMask = 0x81;
    LcdPutPackedX("", 0, (100-12)+32);
    screen.invertMask = 0x00;
    LcdPutGfx("\xff24\x3c00", 2);
    if (locked) { /*keylock symbol*/
	LcdPutGfx("\x384e\x89a9\x894e\x3800", 4);
    } else {
	LcdPutPackedX("", 0, 132);
    }
#endif

#if 1
    //if (cs.playTimeSeconds == 0)
    {
	if (minifatInfo.longFileName[0] == 0) {
	    memsetY(minifatInfo.longFileName, 0,
		    sizeof(minifatInfo.longFileName));
	    memcpyYY(minifatInfo.longFileName, minifatInfo.fileName, 8/2);
	    minifatInfo.longFileName[4] = (' '<<8)|'.';
	    minifatInfo.longFileName[5] = minifatInfo.fileName[8/2];
	    minifatInfo.longFileName[6] = minifatInfo.fileName[8/2+1] & 0xff00U;
	}
	LcdLocate(0, 4);
	LcdPutPackedY(&minifatInfo.longFileName[0], 42/2, 132);
	/* clear the next line in case the old filename was two lines */
	LcdPutPackedX("", 0/2, 132);
    }
#endif
#ifdef GET_PLAYTIME
#if 0
    LcdLocate(0, 6);
    LcdPutLong(lastSample);
    LcdPutChar('/');
    LcdPutNum(cs.sampleRate);
    LcdPutPackedX("", 0/2, 132);
#endif
#if 0 /* total playing time */
    {
	register u_int32 t = lastSample / cs.sampleRate; //cs.playTimeSeconds;
	register u_int16 h = t / 3600;
	LcdLocate(0, 1);
	if (t == -1L) {
	    LcdPutPackedX("\p  ", 1/2, 42);
	} else {
	    LcdPutChar(h%10 +SYMBOL_DIGIT);
	    t -= (u_int32)h * 3600;
	    LcdPutChar(':');
	    LcdPutNum((u_int16)t / 60);
	    LcdPutChar(':');
	    LcdPutNum((u_int16)t % 60);
	}
    }
    LcdPutPackedX("", 0/2, 132);
#endif
#if 1 /* remaining playing time */
    if (cs.playTimeSeconds != -1) {
	register u_int32 t = lastSample / cs.sampleRate - cs.playTimeSeconds;
	register u_int16 h = t / 3600;
	LcdLocate(0, 1);
	LcdPutChar(h%10 +SYMBOL_DIGIT);
	t -= (u_int32)h * 3600;
	LcdPutChar(':');
	LcdPutNum((u_int16)t / 60);
	LcdPutChar(':');
	LcdPutNum((u_int16)t % 60);

	LcdPutPackedX("\p remain", 8/2, 132);
    }
#endif
#endif
}

void MyUserInterfaceIdleHook(void) { /*94 words*/
    if (uiTrigger) {
	PlayerDisplay();

	NandPutCommand(0); /*Bring data bus low*/
	UserInterfaceIdleHook();

#ifdef QUIET_CANCEL
	if (cs.cancel) {
	    volumeReg = 0xc0c0;
	    SetVolume();
	}
#endif
    }
#ifdef BACKLIGHT_DIM
    if (timeCount - backLightTime > BACKLIGHT_DIM) {
	LcdContrast(0x00);
	backLightTime = timeCount;
	SetHookFunction((u_int16)USBSuspend, RealUSBSuspend);
    }
#endif
}

auto void MyPowerOff(void) { /*21 words*/
    LCD_SELECT;
    LCD_SET_COMMAND_MODE;
    SpiSendReceive(0xae);
    BusyWait10();
    SpiSendReceive(0xad); //dcdc off
    SpiSendReceive(0x8a); //dcdc off value
    LCD_END;
    BusyWait10();
//putch('o');
//putch('f');
//putch('f');
    RealPowerOff();
}

#endif/*USE_DISPLAY*/


void main(void) { /*264 words*/
    register u_int32 time;
//putch('d');

#ifdef USE_DISPLAY
    /* Initialize OLED. If you have some other serial LCD,
       use the correct init string. */
    LcdInit(OLED_INIT_STRING);
    LcdClear();
    LcdLocate(0, 1);
    //LcdPutPackedX("\pMLC Player", 10, 132);

    /* Idle hook will update the display */
    SetHookFunction((u_int16)IdleHook, MyUserInterfaceIdleHook);

    /* Turn off display before power off */
    SetHookFunction((u_int16)PowerOff, MyPowerOff);
#ifdef BACKLIGHT_DIM
    SetHookFunction((u_int16)KeyEventHandler, MyKeyEventHandler);

    LcdContrast(0x40);
    backLightTime = timeCount;
    SetHookFunction((u_int16)USBSuspend, NullHook); /* disable suspend */
#endif
#endif

    ph->Read = Large8kPageReadSingle; /* replace physical read function */
    map = FsMapTnCreate(ph, 0); /* Create tiny Mapper to initialize structure*/
    map->Read = FsMlcMapRead1; /* replace mapper read function */
    fsMapTiny.firstBlock = -1; /* invalidate mapping cache */
    fsMapTiny.root =
	((ph->eraseBlockSize < 0x100) ? 0x100 : ph->eraseBlockSize)
	+ ph->eraseBlockSize /*reserve one*/; /* determine disk start */

    /* Patch a decoding problem (when left channel is digital zero)
       plus returns faster after finishing a file. */
    SetHookFunction((u_int16)PlayCurrentFile, PatchPlayCurrentFile);

    clockX = 6; /*3.0x*/
    /* set cs.sampleRate to non-zero because we use cs.Output() */
    cs.sampleRate = 44100U;

    /* Turn off after 5 minutes (300secs) in pause mode. */
    //player.offDelay = 300/5; 
    {
	/* The display needs the IO voltage to be about 2.6V */
	static const __y myVoltages[] = {
	    /* Core I/O Analog */
	    22, 27, 12, /* 2.2 3.3 2.8 player */
	    27, 27, 30, /* 2.5 3.3 3.6 USB */
	    14, 17,  0, /* 1.8 2.6 3.6 (USB) suspend  18=2.65*/
	    22, 27, 12  /* 2.2 3.3 2.8 user -- used for battery measurement*/
	};
	memcpyYX(voltages, myVoltages, sizeof(myVoltages));
	PowerSetVoltages(&voltages[voltCorePlayer]);
    }

#ifndef USE_DISPLAY
    /* Set the leds after nand-boot! */
    PERIP(GPIO1_ODATA) |=  LED1;   /* POWER led on */
    PERIP(GPIO1_DDR)   |=  LED1|LED2;   /* SI and SO to outputs */
    PERIP(GPIO1_MODE)  &= ~(LED1|LED2); /* SI and SO to GPIO */
#endif

    player.volumeOffset = -24;
    player.pauseOn = 0;
    //player.randomOn = 0;
    keyOld = KEY_POWER;
    keyOldTime = -32767; /* ignores the first release of KEY_POWER */

    PlayerVolume();

//    SetHookFunction((u_int16)OpenFile, FatFastOpenFile);
    SetHookFunction((u_int16)OpenFile, FatFastOpenFileD);/*Shorter code*/
    /*** Player loop ***/
    while (1) {
	/* If MMC is not available, try to reinitialize it. */

	/* If USB is attached, call our version of the MassStorage()
	   handling function. */
	if (USBIsAttached()) {
#ifdef USE_DISPLAY
	    LcdClear();
	    LcdLocate(0, 1);
	    LcdPutPackedX("\pUSB Attached", 10, 132);

	    SetHookFunction((u_int16)IdleHook, UserInterfaceIdleHook);
#endif
#if 0
	    /* ReBoot */
	    ph->Read(ph, 0, 16/*max*/, mallocAreaX, minifatBuffer);
#else
	    {
		register int i;
		for(i=0;i<16;i++) {
		    Large8kPageReadSingle(ph, (0>>9) + i, 1,
					  mallocAreaX+i*512/2, minifatBuffer);
		}
	    }
#endif
	    BootFromX(mallocAreaX+8);
	    /* should not return */
	}

	/* Try to init FAT. */
	player.totalFiles = 0;
	if (InitFileSystem() == 0) {
	    /* Restore the default suffixes. */
	    minifatInfo.supportedSuffixes = supportedFiles;
	    player.totalFiles = OpenFile(0xffffU);

	    if (player.totalFiles == 0) {
		/* If no files found, output some samples.
		   This causes the idle hook to be called, which in turn
		   scans the keys and you can turn off the unit if you like.
		*/
#ifdef USE_DEBUG
		puts("no .ogg");
#endif
#ifdef USE_DISPLAY
		LcdLocate(0, 1);
		LcdPutPackedX("\pNo Files", 8/2, 132);
#endif
		goto noFSnorFiles;
	    }

	    /*
	      Currently starts at the first file after a card has
	      been inserted.

	      Possible improvement would be to save play position
	      to SPI EEPROM for a couple of the last-seen MMC/SD's.
	      Could also save the play mode, volume, and earspeaker
	      state to EEPROM.
	    */
//	    player.nextStep = 1;
	    player.nextFile = 0;
	    while (1) {
		player.nextStep = 1;

		if (player.randomOn) {
		    /* Use shuffle from dev1000 library. */
		    player.currentFile =
			Shuffle(player.totalFiles, player.currentFile);
		} else {
		    player.currentFile = player.nextFile;
		}
		if (player.currentFile < 0)
		    player.currentFile += player.totalFiles;
		if (player.currentFile >= player.totalFiles)
		    player.currentFile -= player.totalFiles;
		player.nextFile = player.currentFile + 1;

		/* If the file can be opened, start playing it. */
		memsetY(minifatInfo.longFileName, 0,
			sizeof(minifatInfo.longFileName));
		if (OpenFile(player.currentFile) < 0) {
#ifdef BACKLIGHT_DIM
		    backLightTime = timeCount;
#endif
#ifdef GET_PLAYTIME
		    GetPlayTime();
#endif
		    //player.volumeOffset = -12; /*default volume offset*/
		    //PlayerVolume();
		    player.ffCount = 0;
		    cs.cancel = 0;
		    cs.goTo = -1; /* start playing from the start */
		    cs.fileSize = cs.fileLeft = minifatInfo.fileSize;
		    cs.fastForward = 1; /* reset play speed to normal */
#ifdef QUIET_CANCEL
		    PlayerVolume();
#endif
		    {
			register s_int16 oldStep = player.nextStep;
			register s_int16 ret;
			ret = PlayCurrentFile();
#if 1
			if (earSpeakerReg) {
			    int i;
			    for (i=0;i<2048/32;i++) {
				memset(tmpBuf, 0, sizeof(tmpBuf));
				AudioOutputSamples(tmpBuf, sizeof(tmpBuf)/2);
			    }
			}
#endif
			/* If unsupported, keep skipping */
			if (ret == ceFormatNotFound)
			    player.nextFile = player.currentFile + player.nextStep;
			/* If a supported file found, restore play direction */
			if (ret == ceOk && oldStep == -1)
			    player.nextStep = 1;
		    }
		} else {
		    player.nextFile = 0;
		}
		if (USBIsAttached()) { /* .. or USB is attached. */
		    break;
		}
	    }
	} else {
	    /* If not a valid FAT (perhaps because MMC/SD is not inserted),
	       just send some samples to audio buffer and try again. */
#ifdef USE_DISPLAY
	    LcdLocate(0, 1);
	    LcdPutPackedX("\pNo FAT", 6/2, 132);
#endif
	noFSnorFiles:
	    LoadCheck(&cs, 32); /* decrease or increase clock */
	    memset(tmpBuf, 0, sizeof(tmpBuf));
	    AudioOutputSamples(tmpBuf, sizeof(tmpBuf)/2);
	    /* When no samples fit, calls the user interface
	       -- handles volume control and power-off. */
	}
    }
}
